from abc import abstractmethod

from .base_class import BaseClass


class PreprocessorBase(BaseClass):
    """Preprocessing class.

    Responsible for re-scaling, normalization, cropping, resizing, etc.
    The output of ``add_preprocess`` method will be a genertor that generates a batch of data ``(x_batch, y_batch, sample_weight_batch/data_id_batch)``:

    - ``x_batch`` -> ``(config.batch_size, config.input_h, config.input_w, n_channels)``
    - ``y_batch`` (segmentation) -> ``(config.batch_size, config.input_h, config.input_w, n_classes)``
    - ``y_batch`` (classification) -> ``(config.batch_size, n_classes)`` or ``(config.batch_size,)``
    - ``sample_weight_batch`` (segmentation) -> ``(config.batch_size, config.input_h, config.input_w)``
    - ``sample_weight_batch`` (segmentation/classification) -> ``(config.batch_size,)``
    - ``data_id_batch`` (just for ``test_data_generator``) -> ``(config.batch_size,)``

    Notes:

        - Make sure that your ``Preprocessor`` is compatible with ``Augmentation``, ``DataLoader`` and ``ModelBuilder``.
        - Consider implementing ``image_preprocess`` and ``label_preprocess`` in a way that is compatible with serving using ``SavedModel``.
        - Note that your preprocessing has to be the same in training and inference.
        - Preprocessing will be plugged on top of ``DataLoader``'s generator, after plugging ``Augmentation``.

    Example:
        >>> preprocessor = Preprocessor(config)
        >>> data_gen = preprocessor.add_image_preprocess(data_gen)
        >>> data_gen = preprocessor.add_label_preprocess(data_gen)
        >>> data_gen, n_iter_per_epoch = preprocessor.batchify(data_gen, n_data_points)

    this is the how ``add_preprocess`` method works:

        ``add_image_preprocess`` -> ``add_label_preprocess`` -> ``batchify``

    """

    @abstractmethod
    def image_preprocess(self, image):
        """Image preprocessing logic.

        Args:
            image: input image

        Returns:
            preprocessed image -> ``tf.tensor`` or ``numpy.ndarray`` of ``shape(input_height, input_width, n_channels)``
        """

    @abstractmethod
    def label_preprocess(self, label):
        """Label preprocessing logic.

        Args:
            label: input label

        Returns:
            preprocessed label -> ``tf.tensor`` or ``numpy.ndarray`` of shape

                - segmentation -> ``(input_height, input_width, n_classes)``
                - classification -> ``(n_classes,)`` or scalar ``class_id``
        """

    @abstractmethod
    def add_image_preprocess(self, generator):
        """Plugs input-image-preprocessing on top of the given ``generator``/``tf.Dataset``.

        Notes:
            - you have to do preprocessing on image (``x``), keep label-preprocessing (if any) for ``self.add_label_preprocess``.
            - you have to use exactly the same logic that you define in ``self.image_preprocess`` method. Consider using ``map`` method for ``tf.data.Dataset``, or {``map``function/``for`` loop} for vanilla ``Python generator``.

        Args:
            generator: a ``Python generator``/``tf.data.Dataset`` which yields a single data-point ``(x, y, sample_weight)`` or ``(x, y, data_id)`` if it is ``test_data_generator`` in which

                - ``x`` => input image,
                - ``y`` => label, or segmentation map for segmentation
                - ``sample_weight`` => float (classification/segmentation), or one-channel segmentation map (segmentation)

        Returns:
            A ``generator``/``tf.data.Dataset`` with preprocessed ``x`` s, check ``self.image_preprocess`` for shapes

        """

    @abstractmethod
    def add_label_preprocess(self, generator):
        """Plugs input-label-preprocessing on top of the given ``generator``/``tf.Dataset``.

        Notes:
            - you have to do preprocessing on label (``y``), keep input-image-preprocessing (if any) for ``self.add_image_preprocess``.
            - you have to use exactly the same logic that you define in ``self.label_preprocess`` method. Consider using ``map`` method for ``tf.data.Dataset``, or {``map``function/``for`` loop} for vanilla ``Python generator``.

        Args:
            generator: a ``Python generator``/``tf.data.Dataset`` which yields a single data-point ``(x, y, sample_weight)`` or ``(x, y, data_id)`` if it is ``test_data_generator`` in which

                - ``x`` => input image,
                - ``y`` => label, or segmentation map for segmentation
                - ``sample_weight`` => float (classification/segmentation), or one-channel segmentation map (segmentation)

        Returns:
            A ``generator``/``tf.data.Dataset`` with preprocessed ``y`` s, check ``self.label_preprocess`` for shapes

        """

    @abstractmethod
    def batchify(self, generator, n_data_points):
        """Batchifies the given ``generator``/``tf.Dataset``

        This will be the latest block for data pipeline.

        Notes:
            - you have to return ``n_iter`` in a correct way, consider a last batch of different size if ``batch_size`` does not fit into ``n_data_points``.
            - you can use ``.batch(..).repeat()`` for ``tf.data.Dataset``, or a similar process for ``Python generator``.
            - you can access to ``batch_size`` through ``self.config.batch_size``

        Args:
             generator: a ``Python generator``/``tf.data.Dataset`` which yields a single data-point ``(x, y, sample_weight)`` or ``(x, y, data_id)`` if it is ``test_data_generator`` in which

                - ``x`` =>  (input_h, input_w, n_channels),
                - ``y`` => (n_classes, 1)(classification), or (input_h, input_w, h) for segmentation
                - ``sample_weight`` => float (classification/segmentation), or one-channel segmentation map (segmentation)

            n_data_points: number of data_points in this sub-set.

        Returns:
            tuple(batched_generator, n_iter):
            - batched_generator: A repeated ``generator``/``tf.data.Dataset`` which yields a batch of data for each iteration, ``(x_batch, y_batch, sample_weights)`` or ``(x_batch, y_batch, data_ids)`` for test data gen, in which

                - ``x`` => (batch_size, input_h, input_w, n_channels)
                - ``y`` => classification: (batch_size, n_classes[or 1]), segmentation: (batch_size, input_h, input_w, n_classes)
                - ``sample_weights`` => (batch_size, 1)(classification/segmentation), (batch_size, input_h, input_w, 1)(segmentation)

            - n_iter: number of iterations per epoch

        """

    def add_preprocess(self, generator, n_data_points):
        """Plugs preprocessing on top of given generator.

        Args:
            generator: see ``DataLoaderBase.create_data_generator``
            n_data_points: number of data_points in this sub-set.

        Returns:
            tuple(batched_generator, n_iter):
            - batched_generator: A repeated ``generator``/``tf.data.Dataset`` which yields a batch of data for each iteration, ``(x_batch, y_batch, sample_weights)`` or ``(x_batch, y_batch, data_ids)`` for test data gen, in which

                - ``x`` => (batch_size, input_h, input_w, n_channels)
                - ``y`` => classification: (batch_size, n_classes[or 1]), segmentation: (batch_size, input_h, input_w, n_classes)
                - ``sample_weights`` => (batch_size, 1)(classification/segmentation), (batch_size, input_h, input_w, 1)(segmentation)

            - n_iter: number of iterations per epoch

        """

        gen = self.add_image_preprocess(generator)
        gen = self.add_label_preprocess(gen)
        gen, n_iter = self.batchify(gen, n_data_points)
        return gen, n_iter

    @classmethod
    def __subclasshook__(cls, c):
        """This defines the __subclasshook__ class method.
        This special method is called by the Python interpreter to answer the question,
         Is the class C a subclass of this class?
        """

        if cls is PreprocessorBase:
            attrs = set(dir(c))
            if set(cls.__abstractmethods__) <= attrs:
                return True
        return NotImplemented



import types
from abc import abstractmethod

import mlflow
from tqdm import tqdm
import typing

import numpy as np
import pandas as pd
import tensorflow.keras as tfk
import tensorflow as tf

from .base_class import BaseClass
from .data_loading import DataLoaderBase
from .preprocessing import PreprocessorBase


EvalFuncs = typing.Dict[str, types.FunctionType]


class EvaluatorBase(BaseClass):

    def _load_params(self, config):
        pass

    def _set_defaults(self):
        pass

    def evaluate(self,
                 data_loader: DataLoaderBase,
                 preprocessor: PreprocessorBase,
                 exported_model: tfk.Model,
                 active_run: mlflow.ActiveRun) -> pd.DataFrame:
        """Evaluates the model using ``eval_functions`` defined in ``get_eval_functions`` on test dataset.

        Notes:
            - ``create_test_generator`` method of the ``data_loader`` will be used to create a data generator, and this method expects a tuple of ``(image, label, data_id)`` as output.

        Args:
            data_loader: will be used for creating test-data-generator
            preprocessor: will be used to add image-preprocess to the test-data-generator
            exported_model: a ``tensorflow.keras.Model`` which is ready for making inference.
            active_run: mlflow's ``ActiveRun`` instance to log evaluation reports on.

        Returns:
            a ``pandas.DataFrame`` with index= ``data_id`` s, each row represents the result for a single data-point and
            each column represents a metric.

        """

        test_data_gen, test_n = data_loader.create_test_generator()
        test_data_gen = preprocessor.add_image_preprocess(test_data_gen)
        test_data_gen = preprocessor.add_label_preprocess(test_data_gen)

        report_df = self._get_eval_report(test_data_gen, test_n, exported_model)
        self._log_to_mlflow(active_run, report_df)

        return report_df

    def validation_evaluate(self,
                            data_loader: DataLoaderBase,
                            preprocessor: PreprocessorBase,
                            exported_model: tfk.Model,
                            active_run: mlflow.ActiveRun,
                            index) -> pd.DataFrame:
        """Evaluates the model using ``eval_functions`` defined in ``get_eval_functions`` on validation dataset.

        Notes:
            - ``create_validation_generator`` method of the ``data_loader`` will be used to create a data generator, and this method expects a tuple of ``(image, label, data_id/sample_weight)`` as output.

        Args:
            data_loader: will be used for creating test-data-generator.
            preprocessor: will be used to add image-preprocess to the test-data-generator.
            exported_model: a ``tensorflow.keras.Model`` which is ready for making inference.
            active_run: mlflow's ``ActiveRun`` instance to log evaluation reports to.
            index: a list of ``data_id`` s to use as report data-frame's index. use this if your validation generator does not return a unique ``data_id`` as third element.

        Returns:
            a ``pandas.DataFrame`` with index= ``data_id`` s, each row represents the result for a single data-point
            and each column represents a metric.

        """

        val_data_gen, val_n = data_loader.create_validation_generator()
        val_data_gen = preprocessor.add_image_preprocess(val_data_gen)

        report_df = self._get_eval_report(val_data_gen, val_n, exported_model)
        self._log_to_mlflow(active_run, report_df, prefix='validation')
        if index is not None:
            report_df.index = index
        else:
            report_df.index = list(range(val_n))

        return report_df

    @staticmethod
    def _wrap_pred_step(model):
        """Overrides the ``predict`` method of the model.

        By calling ``predict`` method of the model, three lists will be returned:
         ``(predictions, ground truths, data_ids/sample_weights)``
        """

        def new_predict_step(data):
            x, y, z = tfk.utils.unpack_x_y_sample_weight(data)
            return model(x, training=False), y, z

        setattr(model, 'predict_step', new_predict_step)

    @staticmethod
    def _log_to_mlflow(active_run: mlflow.ActiveRun, report_df: pd.DataFrame, prefix='test'):
        summary_report = report_df.describe()

        test_metrics = {}
        for c in summary_report.columns:
            metric_name = f'{prefix}_{c}'
            metric_value = summary_report[c]['mean']
            test_metrics[metric_name] = metric_value

        with active_run:
            mlflow.set_tag("session_type", "evaluation")
            mlflow.log_metrics(test_metrics)

    def _get_eval_report(self, data_gen, data_n, model):
        eval_funcs = self.get_eval_funcs()
        report = {k: list() for k in eval_funcs.keys()}
        indxs = list()
        count = 0
        with tqdm(total=data_n) as pbar:
            try:
                for elem in data_gen:
                    data_id = elem[2]
                    y_true = elem[1]
                    x = elem[0]
                    y_pred = model.predict(np.expand_dims(x, axis=0))[0]

                    if isinstance(data_id, tf.Tensor):
                        indxs.append(str(data_id.numpy()))
                    else:
                        indxs.append(str(data_id))

                    for k, v in report.items():
                        v.append(eval_funcs[k](y_true, y_pred))

                    pbar.update(1)
                    count += 1
                    if count >= data_n:
                        break
            except (StopIteration, RuntimeError):
                print('stop iteration ...')

        df = pd.DataFrame(report, index=indxs)
        return df

    def _batch_eval_report(self, data_gen, n_iter, model):
        #TODO: write this method for performance improvements
        pass

    @abstractmethod
    def get_eval_funcs(self) -> EvalFuncs:
        """Evaluation functions to use for evaluation.

        You should take ``model.predict`` outputs into account. Your functions must take ``(y_true, y_pred)`` as
        inputs (not batch), and return a single output for this data-point.

        Returns:
            a dictionary mapping from metric names to metric functions: ``{'iou': iou_metric, 'dice': get_dice_metric(), ...}``

        """



import yaml
import pathlib
import logging
import sys

import mlflow

from .abs_exceptions import ConfigNotFound, FoundMultipleConfigs


def setup_mlflow(mlflow_tracking_uri, mlflow_experiment_name: str,
                 base_dir: pathlib.Path, evaluation=False) -> mlflow.ActiveRun:
    """Sets up mlflow and returns an ``active_run`` object.

    tracking_uri/
        experiment_id/
            run1
            run2
            ...

    Args:
        mlflow_tracking_uri: ``tracking_uri`` for mlflow
        mlflow_experiment_name: ``experiment_name`` for mlflow, use the same ``experiment_name`` for all experiments
        related to the same task. This is different from the ``experiment`` concept that we use.
        base_dir: directory for your experiment, containing your `config.yaml` file.
        evaluation: if evaluation==true, then new run will be created, named ``base_dir.name + _evaluation``

    Returns:
        active_run: an ``active_run`` object to use for mlflow logging.

    """

    # Loads run_id if exists
    run_id = None
    run_id_path = base_dir.joinpath('run_id.txt')
    run_name = base_dir.name
    if evaluation:
        run_name += '_evaluation'
    elif run_id_path.exists():
        with open(run_id_path, 'r') as f:
            run_id = f.readline()

    mlflow.set_tracking_uri(mlflow_tracking_uri)
    client = mlflow.tracking.MlflowClient(mlflow_tracking_uri)

    # Create new run if run_id does not exist
    if run_id is not None:
        mlflow.set_experiment(mlflow_experiment_name)
        active_run = mlflow.start_run(run_id=run_id)
    else:
        experiment = client.get_experiment_by_name(mlflow_experiment_name)
        if experiment is not None:
            experiment_id = experiment.experiment_id
        else:
            experiment_id = mlflow.create_experiment(mlflow_experiment_name)

        active_run = mlflow.start_run(experiment_id=experiment_id, run_name=run_name)

    return active_run


def check_for_config_file(run_dir: pathlib.Path) -> pathlib.Path:
    """Checks for existence of config file and returns the path to config file if exists.

    Raises:
        ConfigNotFound
        FoundMultipleConfigs

    Returns:
        path to config file

    """

    if not run_dir.is_dir():
        raise Exception(f'{run_dir} is not a directory.')

    yaml_files = list(run_dir.glob('*.yaml'))
    if not any(yaml_files):
        raise ConfigNotFound(f'no .yaml files found.')
    elif len(yaml_files) > 1:
        raise FoundMultipleConfigs(f'found more than one .yaml files.')

    return yaml_files[0]


class ConfigStruct:
    """Structure for loading config as a Python object.

    Attributes:
        seed (int):
        input_height (int):
        input_width (int):
        src_code_path (str):
        data_loader_class (str):
        model_builder_class (str):
        preprocessor_class (str):
        augmentor_class (str):
        evaluator_class (str):
        epochs (int):
        batch_size (int):
        data_loader (Struct):
        model_builder (Struct):
        preprocessor (Struct):
        augmentor (Struct):
        do_train_augmentation (bool):
        do_validation_augmentation (bool):
        export (Struct):
        project_name (str):

    """

    def __init__(self, **entries):
        self.seed = 101
        self.input_height = None
        self.input_width = None
        self.src_code_path = None
        self.data_loader_class = None
        self.model_builder_class = None
        self.preprocessor_class = None
        self.augmentor_class = None
        self.evaluator_class = None
        self.epochs = None
        self.batch_size = None
        self.data_loader = None
        self.model_builder = None
        self.preprocessor = None
        self.augmentor = None
        self.do_train_augmentation = None
        self.do_validation_augmentation = None
        self.export = None
        self.project_name = None

        for k, v in entries.items():
            if isinstance(v, dict):
                self.__dict__[k] = Struct(**v)
            else:
                self.__dict__[k] = v


class Struct:
    def __init__(self, **entries):
        for k, v in entries.items():
            if isinstance(v, dict):
                self.__dict__[k] = Struct(**v)
            else:
                self.__dict__[k] = v


def load_config_file(path: pathlib.Path) -> ConfigStruct:
    """
    loads the json config file and returns a dictionary

    Args:
        path: path to json config file

    Returns:
        a nested object in which parameters are accessible using dot notations, for example ``config.model.optimizer.lr``

    """

    with open(path) as f:
        data_map = yaml.safe_load(f)

    config_obj = ConfigStruct(**data_map)
    return config_obj


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    # if write_logs:
    #     dir_path = os.path.dirname(os.path.realpath(__file__))
    #     log_dir = os.path.join(dir_path, 'logs')
    #     if not os.path.exists(log_dir):
    #         os.mkdir(log_dir)
    #     file_handler = logging.FileHandler(os.path.join(log_dir, "{}.log".format(name)))
    #     file_handler.setFormatter(formatter)
    #     logger.addHandler(file_handler)
    logger.setLevel(logging.INFO)
    return logger



from abc import ABC, abstractmethod

from .utils import ConfigStruct


class BaseClass(ABC):

    def __init__(self, config: ConfigStruct = None):
        self._set_defaults()
        self.config = config
        if config is not None:
            self._load_params(config)

    @abstractmethod
    def _load_params(self, config):
        """Read parameters from config file."""

        pass

    @abstractmethod
    def _set_defaults(self):
        """Default values for your class, if ``None`` is passed as config.

        Should initialize the same parameters as in ``_load_params`` method.
        """

        pass



from pathlib import Path

MLFLOW_TRACKING_URI = Path('/home').joinpath('vafaeisa').joinpath('mlruns')
scratch_drive = Path('/home').joinpath('vafaeisa').joinpath('scratch')
EVAL_REPORTS_DIR = scratch_drive.joinpath('eval_reports')
PROJECT_ROOT = Path(__file__).parent.parent

from .base_class import BaseClass
from .model_building import ModelBuilderBase
from .data_loading import DataLoaderBase
from .evaluation import EvaluatorBase
from .preprocessing import PreprocessorBase
from .augmentation import AugmentorBase
from .training import Trainer
from .orchestration import Orchestrator



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='abstractions-aimedic',
    version='0.1.5',
    license='MIT',
    author='Soroush Moazed',
    author_email='soroush.moazed@gmail.com',
    description="Abstractions for AIMedic's training components.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.aimedic.co/soroush.moazed/abstractions.git",
    download_url="https://github.com/iamsoroush/abstractions/archive/refs/tags/v_0.1.4.tar.gz",
    classifiers=[
            'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
            'Intended Audience :: Developers',      # Define that your audience are developers
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',   # Again, pick a license
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[
        'pyyaml',
        'tensorflow>=2.5',
        'tqdm',
        'pandas',
        'mlflow'
    ],
    # packages=['abstractions'],
    python_requires=">=3.6"
)


import os
import pathlib
import yaml
import shutil

import tensorflow.keras as tfk
import tensorflow.keras.callbacks as tfkc
import mlflow

from .base_class import BaseClass
from .model_building import ModelBuilderBase
from .abs_exceptions import *


class Trainer(BaseClass):

    def _load_params(self, config):
        self.epochs = config.epochs
        self.export_metric = config.export.metric
        self.export_mode = config.export.mode

    def _set_defaults(self):
        self.epochs = 10
        self.export_metric = 'val_loss'
        self.export_mode = 'min'

    def __init__(self, config, run_dir):
        self.run_dir = pathlib.Path(run_dir)
        super().__init__(config=config)

        # Paths
        self.checkpoints_dir = self.run_dir.joinpath('checkpoints')
        self.tensorboard_log_dir = self.run_dir.joinpath('logs')
        self.run_id_path = self.run_dir.joinpath('run_id.txt')
        self.exported_dir = self.run_dir.joinpath('exported')
        self.exported_saved_model_path = self.exported_dir.joinpath('savedmodel')
        self.config_file_path = [i for i in self.run_dir.iterdir() if i.name.endswith('.yaml')][0]

        # Raise exception if the best model is already exported.
        self._check_for_exported()

        # Make logs and checkpoints directories
        self.checkpoints_dir.mkdir(exist_ok=True)
        self.tensorboard_log_dir.mkdir(exist_ok=True)

        # Container for fit_history
        self.history_ = None

    def train(self,
              model_builder: ModelBuilderBase,
              active_run: mlflow.ActiveRun,
              train_data_gen,
              n_iter_train,
              val_data_gen,
              n_iter_val):
        """Trains the model using data generators and logs to ``mlflow``.

        Will try to resume training from latest checkpoint, else starts training from ``epoch=0``.

        Args:
            model_builder: will be used for generating model if no checkpoints could be find.
            active_run: ``mlflow.ActiveRun`` for logging to **mlflow**. This will be used inside a context manager.
            train_data_gen: preprocessed, augmented training-data-generator.This will be the output of
             ``Preprocessor.add_batch_preprocess``.
            n_iter_train: ``steps_per_epoch`` for ``model.fit``
            val_data_gen: preprocessed, augmented validation data-generator.This will be the output of
             ``Preprocessor.add_batch_preprocess``.
            n_iter_val: ``validation_steps`` for ``model.fit``

        """

        # Create/load model and define initial_epoch
        if any(self._get_checkpoints()):
            model, initial_epoch = self._load_latest_model()
        else:
            initial_epoch = 0
            model = model_builder.get_compiled_model()

        # Get callbacks
        callbacks = self._get_callbacks(model_builder)

        # Run
        with active_run as run:
            # Add params from config file to mlflow
            self._add_config_file_to_mlflow()

            # Set dataset name as a tag
            mlflow.set_tag("session_type", "training")  # ['hpo', 'evaluation', 'training']

            mlflow.tensorflow.autolog(every_n_iter=1,
                                      log_models=False,
                                      disable=False,
                                      exclusive=False,
                                      disable_for_unsupported_versions=False,
                                      silent=False)

            # Write run_id
            self._write_mlflow_run_id(run)

            # Fit
            self.history_ = model.fit(train_data_gen,
                                      steps_per_epoch=n_iter_train,
                                      initial_epoch=initial_epoch,
                                      epochs=self.epochs,
                                      validation_data=val_data_gen,
                                      validation_steps=n_iter_val,
                                      class_weight=model_builder.get_class_weight(),
                                      callbacks=callbacks)

    def export(self) -> pathlib.Path:
        """Exports the best version of ``SavedModel`` s, and ``config.yaml`` file into exported sub_directory.

        This method will delete all checkpoints after exporting the best one.

        Returns:
            directory to exported model and config file.

        """

        best_model_info = self._get_best_checkpoint()

        # exported_saved_model_path = self.exported_dir.joinpath('savedmodel')
        exported_config_path = self.exported_dir.joinpath('config.yaml')
        shutil.copytree(best_model_info['path'], self.exported_saved_model_path,
                        symlinks=False, ignore=None, ignore_dangling_symlinks=False)
        shutil.copy(self.config_file_path, exported_config_path)

        # Delete checkpoints
        shutil.rmtree(self.checkpoints_dir)
        return self.exported_dir

    def _get_callbacks(self, model_builder: ModelBuilderBase):
        """Makes sure that TensorBoard and ModelCheckpoint callbacks exist and are correctly configured.

        Attributes:
            model_builder: ``ModelBuilder`` object, to get callbacks list using ``model_builder.get_callbacks``

        modifies ``callbacks`` to be a list of callbacks, in which ``TensorBoard`` callback exists with
         ``log_dir=self.tensorboard_log_dir`` and ``ModelCheckpoint`` callback exists with
          ``filepath=self.checkpoints_dir/...``, ``save_weights_only=False``

        """

        callbacks = model_builder.get_callbacks()

        mc_callbacks = [i for i in callbacks if isinstance(i, tfkc.ModelCheckpoint)]
        tb_callbacks = [i for i in callbacks if isinstance(i, tfkc.TensorBoard)]

        to_track = self.export_metric
        checkpoint_path = str(self.checkpoints_dir) + "/sm-{epoch:04d}"
        checkpoint_path = checkpoint_path + "-{" + to_track + ":4.5f}"

        if any(mc_callbacks):
            mc_callbacks[0].filepath = str(checkpoint_path)
            mc_callbacks[0].save_weights_only = False
        else:
            mc = tfkc.ModelCheckpoint(filepath=checkpoint_path,
                                      save_weights_only=False)
            callbacks.append(mc)

        if any(tb_callbacks):
            tb_callbacks[0].log_dir = self.tensorboard_log_dir
        else:
            tb = tfkc.TensorBoard(log_dir=self.tensorboard_log_dir)
            callbacks.append(tb)

        return callbacks

    def _load_latest_model(self):
        """Loads and returns the latest ``SavedModel``.

        Returns:
            model: a ``tf.keras.Model`` object.
            initial_epoch: initial epoch for this checkpoint

        """

        latest_ch = self._get_latest_checkpoint()
        initial_epoch = latest_ch['epoch']
        sm_path = latest_ch['path']
        print(f'found latest checkpoint: {sm_path}')
        print(f'resuming from epoch {initial_epoch}')
        model = tfk.models.load_model(latest_ch['path'])
        return model, initial_epoch

    def _check_for_exported(self):
        """Raises exception if exported directory exists and contains some files"""

        if self.exported_dir.is_dir():
            if any(self.exported_dir.iterdir()):
                raise ExportedExists('exported files already exist.')

    def _get_latest_checkpoint(self):
        """Returns info about the latest checkpoint.

        Returns:
            a dictionary containing epoch, path to ``SavedModel`` and value of ``self.export_metric`` for
            latest checkpoint:
                {'epoch': int, 'path': pathlib.Path, 'value': float}

        """

        checkpoints = self._get_checkpoints_info()
        return max(checkpoints, key=lambda x: os.path.getctime(x['path']))

    def _get_best_checkpoint(self):
        """Returns info about the best checkpoint.

        Returns:
            a dictionary containing epoch, path to ``SavedModel`` and value of ``self.export_metric`` for
            the best checkpoint in terms of ``self.export_metric``:
                {'epoch': int, 'path': pathlib.Path, 'value': float}

        """

        checkpoints = self._get_checkpoints_info()

        if self.export_mode == 'min':
            selected_model = min(checkpoints, key=lambda x: x['value'])
        else:
            selected_model = max(checkpoints, key=lambda x: x['value'])
        return selected_model

    def _get_checkpoints(self):
        """Returns a list of paths to folders containing a ``saved_model.pb``"""

        ckpts = [item for item in self.checkpoints_dir.iterdir() if any(item.glob('saved_model.pb'))]
        return ckpts

    def _get_checkpoints_info(self):
        """Returns info about checkpoints.

        Returns:
            A list of dictionaries related to each checkpoint:
                {'epoch': int, 'path': pathlib.Path, 'value': float}

        """

        checkpoints = self._get_checkpoints()
        ckpt_info = list()
        for cp in checkpoints:
            splits = str(cp.name).split('-')
            epoch = int(splits[1])
            metric_value = float(splits[2])
            ckpt_info.append({'path': cp, 'epoch': epoch, 'value': metric_value})
        return ckpt_info

    def _add_config_file_to_mlflow(self):
        """Adds parameters from config file to mlflow.

        Be sure to call this function inside a context manager using a ``mlflow.ActiveRun``.

        """

        def param_extractor(dictionary):

            """Returns a list of each item formatted like 'trainer.mlflow.tracking_uri: /tracking/uri' """

            values = []
            if dictionary is None:
                return values

            for key, value in dictionary.items():
                if isinstance(value, dict):
                    items_list = param_extractor(value)
                    for i in items_list:
                        values.append(f'{key}.{i}')
                else:
                    values.append(f'{key}: {value}')
            return values

        with open(self.config_file_path) as file:
            data_map = yaml.safe_load(file)

        str_params = param_extractor(data_map)
        params = {}
        for item in str_params:
            name = item.split(':')[0]
            item_value = item.split(': ')[-1]
            params[name] = item_value

        mlflow.log_params(params)

    def _write_mlflow_run_id(self, run: mlflow.ActiveRun):
        with open(self.run_id_path, 'w') as f:
            f.write(run.info.run_id)



class ExportedExists(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg


class ConfigNotFound(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg


class FoundMultipleConfigs(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg


class ConfigParamDoesNotExist(Exception):
    pass



from abc import abstractmethod

from .base_class import BaseClass


class AugmentorBase(BaseClass):
    """Augmentor class.

    Notes:
        - Augmentations should not change the input image/label's d-type, size and range, any modifications that need to be the same in inference adn training must be implemented in ``Preprocessor``
        - Make sure that label/image augmentation are compatible.

    Examples:
        >>> augmentor = Augmentor(config)
        >>> data_gen = augmentor.add_augmentation(data_gen)

    """

    @abstractmethod
    def add_augmentation(self, generator):
        """Plugs augmentation to the end of the given ``generator``/``tf.Dataset``.

        See ``DataLoader`` for more info about the input generator.

        Args:
            generator: a ``Python generator`` or ``tf.data.Dataset`` which yields a single data-point of
             ``(x, y, sample_weight(s))`` or ``(x, y, data_id)`` if it is ``test_data_generator``.

                - ``x``: input image,
                - ``y``: label, or segmentation map (segmentation)
                - ``sample_weight``: float (classification/segmentation), or binary segmentation map (segmentation)

        Returns:
            A ``generator``/``tf.data.Dataset`` with augmented elements.

        """

    @abstractmethod
    def add_validation_augmentation(self, generator):
        """If you have different augmentation strategy for validation.

        If you don't want to ``do_validation_augmentation`` or you have similar augmentation strategies for train
        and validation, just skip over-riding this method.

        Args:
            generator: a ``Python generator`` or ``tf.data.Dataset`` which yields a single data-point of
             ``(x, y, sample_weight(s))`` or ``(x, y, data_id)`` if it is ``test_data_generator``.

                - ``x``: input image,
                - ``y``: label, or segmentation map (segmentation)
                - ``sample_weight``: float (classification/segmentation), or binary segmentation map (segmentation)

        Returns:
            A ``generator``/``tf.data.Dataset`` with augmented elements.

        """

        return self.add_augmentation(generator)

    @classmethod
    def __subclasshook__(cls, c):
        """This defines the __subclasshook__ class method.
        This special method is called by the Python interpreter to answer the question,
         Is the class C a subclass of this class?
        """

        if cls is AugmentorBase:
            attrs = set(dir(c))
            if set(cls.__abstractmethods__) <= attrs:
                return True
        return NotImplemented
